from __future__ import annotations

import json
from dbus_next import Variant
from dbus_next.aio import MessageBus
from dbus_next.introspection import Node

from .service import StatusService

notifications_xml = '''
<node>
    <interface name="org.freedesktop.Notifications">
        <method name="GetCapabilities">
            <arg type="as" name="capabilities" direction="out">
            </arg>
        </method>
        <method name="Notify">
            <arg type="s" name="app_name" direction="in">
            </arg>
            <arg type="u" name="replaces_id" direction="in">
            </arg>
            <arg type="s" name="app_icon" direction="in">
            </arg>
            <arg type="s" name="summary" direction="in">
            </arg>
            <arg type="s" name="body" direction="in">
            </arg>
            <arg type="as" name="actions" direction="in">
            </arg>
            <arg type="a{sv}" name="hints" direction="in">
            </arg>
            <arg type="i" name="expire_timeout" direction="in">
            </arg>
            <arg type="u" name="id" direction="out">
            </arg>
        </method>
        <method name="CloseNotification">
            <arg type="u" name="id" direction="in">
            </arg>
        </method>
        <method name="GetServerInformation">
            <arg type="s" name="name" direction="out">
            </arg>
            <arg type="s" name="vendor" direction="out">
            </arg>
            <arg type="s" name="version" direction="out">
            </arg>
            <arg type="s" name="spec_version" direction="out">
            </arg>
        </method>
        <signal name="NotificationClosed">
            <arg type="u" name="id">
            </arg>
            <arg type="u" name="reason">
            </arg>
        </signal>
        <signal name="ActionInvoked">
            <arg type="u" name="id">
            </arg>
            <arg type="s" name="action_key">
            </arg>
        </signal>
    </interface>
</node>
'''


class Block:
    def __init__(self, name):
        self.name = name
        self.notifications = None

    async def connect(self) -> Block:
        bus = await MessageBus().connect()
        obj = bus.get_proxy_object('org.freedesktop.Notifications',
                                   '/org/freedesktop/Notifications', notifications_xml)
        self.notifications = obj.get_interface('org.freedesktop.Notifications')
        obj = bus.get_proxy_object('com.dubstepdish.i3dstatus', '/com/dubstepdish/i3dstatus',
                                   Node(interfaces=[StatusService().introspect()]))
        self.i3dstatus = obj.get_interface('com.dubstepdish.i3dstatus')
        config_json = await self.i3dstatus.call_get_config(self.name)
        self.config = json.loads(config_json)
        return self

    @staticmethod
    def expand_template(text, context):
        if not context:
            return text

        for key in sorted(context.keys(), key=lambda k: len(k), reverse=True):
            text = text.replace('%' + key, str(context[key]))

        return text

    async def clear(self, instance=None):
        block = {
            'name': Variant('s', self.name),
            'full_text': Variant('s', ''),
        }
        if instance:
            block['instance'] = instance

        await self.i3dstatus.call_show_block(block)

    async def show(self, full_text, instance=None, markup=None, context=None):
        block = {
            'name': Variant('s', self.name),
            'full_text': Variant('s', Block.expand_template(full_text, context)),
        }

        if markup is True:
            markup = "pango"
        if markup:
            block['markup'] = Variant('s', markup)
        if instance:
            block['instance'] = Variant('s', instance)

        await self.i3dstatus.call_show_block(block)

    async def notify(self, message):
        if self.notifications:
            # https://developer.gnome.org/notification-spec/
            message = 'i3-dstatus [{generator}]: {msg}'.format(generator=self.name, msg=message)
            await self.notifications.call_notify('i3dstatus', 0, '', '', message, [], {}, -1)

    async def error(self, message):
        # TODO error log
        await self.notify(message)



#!/usr/bin/env python3

import json
import sys
import os
import yaml
import argparse
from argparse import ArgumentParser, RawDescriptionHelpFormatter
import datetime
import traceback
from enum import Enum

from dbus_next.service import ServiceInterface, method
from dbus_next.aio import MessageBus
import asyncio
from asyncio.subprocess import PIPE


class LogLevel(Enum):
    DEBUG = 1
    INFO = 2
    ERROR = 3


class Generator:
    def __init__(self, generator_path):
        self.generator_path = generator_path
        self.proc = None

    async def start(self):
        # XXX: put i3dstatus in the env so generators can use the library in
        # development without a system install (is there a better way?)
        environ = dict(os.environ)
        pythonpath = os.path.dirname(__file__) + '/..'
        if 'PYTHONPATH' in environ:
            pythonpath = environ['PYTHONPATH'] + ':' + pythonpath

        environ['PYTHONPATH'] = pythonpath

        self.proc = await asyncio.subprocess.create_subprocess_exec(self.generator_path,
                                                                    env=environ,
                                                                    stdout=PIPE,
                                                                    stderr=PIPE)
        await self.proc.wait()


class StatusService(ServiceInterface):
    def __init__(self):
        super().__init__('com.dubstepdish.i3dstatus')

    def start(self, generators, stream=sys.stdout, config={}):
        self.blocks = []
        self.generators = generators
        self.config = config if config else {}
        self.stream = stream

        script_dir = os.path.dirname(__file__)

        if 'general' in self.config:
            if 'generators' in self.config['general']:
                # append the generators in the config to the list of generators
                for generator in self.config['general']['generators']:
                    if generator not in self.generators:
                        self.generators.append(generator)

        paths = []
        # arguments are the names of generators to run
        for generator in generators:
            generator_path = ''
            if os.path.isabs(os.path.expanduser(generator)):
                generator_path = os.path.expanduser(generator)
            else:
                generator_path = os.path.join(script_dir, 'generators', generator)

            if os.path.isfile(generator_path):
                paths.append(generator_path)
            else:
                sys.stderr.write("Could not find generator: {}".format(generator))

        self.stream.write('{"version":1}\n[\n[]\n')
        # self.stream..write('{"version":1, "click_events":true}\n[\n[]\n')
        self.stream.flush()

        for generator_path in paths:
            asyncio.ensure_future(Generator(generator_path).start())

        if 'general' in self.config:
            if 'order' in self.config['general']:
                # `order` in the config overrides generator order given on the
                # command line
                self.generators = self.config['general']['order']

    @method()
    def show_block(self, block: 'a{sv}'):
        # apply config options to the block
        block_config = {}

        for k, v in block.items():
            block[k] = v.value

        if block['name'] in self.config:
            if 'instance' in block and \
                    block['instance'] in self.config[block['name']]:
                # this block is configured by instance
                block_config = self.config[block['name']][block['instance']]
            else:
                block_config = self.config[block['name']]

        for i3bar_key in [
                'color', 'min_width', 'min-width'
                'align', 'separator', 'separator_block_width', 'separator-block-width'
        ]:
            if i3bar_key.replace('-', '_') in block:
                # if given in the block, it overrides the config
                continue
            if 'general' in self.config and \
                    i3bar_key in self.config['general']:
                # if given in the general config, this is the default for all
                # blocks
                k = i3bar_key.replace('-', '_')
                block[k] = self.config['general'][i3bar_key]
            if i3bar_key in block_config:
                # if given in the block config, this overrides the general
                # config
                k = i3bar_key.replace('-', '_')
                block[k] = block_config[i3bar_key]

        # try to find the block given by the name and replace it
        block_found = False
        for i, b in enumerate(self.blocks):
            if b['name'] == block['name']:
                block_found = True
                self.blocks[i] = block

        if not block_found:
            self.blocks.append(block)

        # filter out blocks with no 'full_text' member
        self.blocks = [b for b in self.blocks if 'full_text' in b and b['full_text']]

        # sort by the order the generators were given
        def sort_blocks(b):
            if b['name'] in self.generators:
                return self.generators.index(b['name']) + 1
            else:
                return 0

        self.blocks.sort(key=sort_blocks)

        self.stream.write(',' + json.dumps(self.blocks, ensure_ascii=False) + '\n')
        self.stream.flush()

    @method()
    def get_config(self, block_name: 's') -> 's':
        if block_name in self.config:
            return json.dumps(self.config[block_name], ensure_ascii=False)
        else:
            return '{}'

    @method()
    def generator_log(self, log_level: 'i', block_name: 's', message: 's'):
        print('got log message: log_level = {}, block_name = {}, message = {}'.format(
            log_level, block_name, message),
              file=sys.stderr)


async def start():
    description = '''\
Another great statusline generator for i3wm.\
    '''

    epilog = '''\
available generators:
  battery           - show battery information
  check-http        - poll a site and show if it is up or down
  clipboard         - show the contents of the clipboard
  clock             - show the current time
  disk              - show disk usage statistics
  focused-window    - show the currently focused window
  github-repos      - show issues and stars about github repos for a user
  netifaces         - show information about a network interface
  playerctl         - use playerctl to show track metadata
  scratchpad        - show what windows are in the i3 scratchpad

example usage:
  i3-dstatus focused-window ethernet clock
    '''

    parser = ArgumentParser(prog='i3-dstatus',
                            description=description,
                            epilog=epilog,
                            formatter_class=RawDescriptionHelpFormatter)

    parser.add_argument('-c',
                        metavar='CONFIG',
                        dest='config',
                        nargs=1,
                        help='The location of your i3-dstatus config')

    parser.add_argument('generators',
                        metavar='Generators',
                        nargs=argparse.REMAINDER,
                        help='''Pass a list of generators as arguments to
                        appear in the statusline. See below for a list of
                        generators.''')

    args = parser.parse_args()

    # load the config
    config = {}

    try:
        config_path = args.config[0] if args.config else '~/.i3-dstatus.conf'
        with open(os.path.expanduser(config_path)) as f:
            config = yaml.safe_load(f)
    except FileNotFoundError:
        if args.config:
            message = 'Could not find config at {}\n'.format(args.config[0])
            parser.exit(status=1, message=message)
        else:
            pass

    try:
        bus = await MessageBus().connect()
        await bus.request_name('com.dubstepdish.i3dstatus')

        service = StatusService()
        service.start(args.generators, config=config)
        bus.export('/com/dubstepdish/i3dstatus', service)
    except Exception as e:
        with open('/tmp/i3-dstatus-error.log', 'a') as f:
            f.write("ERROR - " + str(datetime.datetime.now()) + "\n")
            f.write(traceback.format_exc())
            f.write('\n')
        raise e

    loop = asyncio.get_event_loop()
    await loop.create_future()


if __name__ == '__main__':
    loop = asyncio.get_event_loop()
    loop.run_until_complete(start())



from setuptools import setup, find_packages

setup(
    name='i3-dstatus',
    version='0.1.1',
    author='Tony Crisci',
    author_email='tony@dubstepdish.com',
    url='https://github.com/altdesktop/i3-dstatus',
    license='BSD',
    description='The ultimate DIY statusline generator for i3',
    long_description=open('README.rst').read(),

    install_requires=['PyYAML', 'netifaces', 'i3ipc', 'requests', 'dbus-next'],

    classifiers=[
        'Development Status :: 4 - Beta',
        'License :: OSI Approved :: BSD License',
        'Programming Language :: Python :: 3',
    ],

    package_data={'i3dstatus': ['generators/*']},

    scripts=['i3-dstatus'],

    entry_points={
        'console_scripts': [
            'i3-dstatus = i3dstatus.__main__:main'
            ]
        },

    packages=find_packages(),
)
